@@ -1,10 +1,10 @@
 /*
- * $Header: /cvsroot/pgsql-server/contrib/pgbench/pgbench.c,v 1.27 2003/09/27 19:15:34 wieck Exp $
+ * $PostgreSQL: pgsql/contrib/pgbench/pgbench.c,v 1.35 2004/11/09 06:09:31 neilc Exp $
  *
  * pgbench: a simple TPC-B like benchmark program for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2000-2002	Tatsuo Ishii
+ * Copyright (c) 2000-2004	Tatsuo Ishii
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -44,6 +44,10 @@
 extern char *optarg;
 extern int	optind;
 
+#ifdef WIN32
+#undef select
+#endif
+
 
 /********************************************************************
  * some configurable parameters */
@@ -103,7 +107,7 @@ typedef struct
 }	CState;
 
 static void
-usage()
+usage(void)
 {
 	fprintf(stderr, "usage: pgbench [-h hostname][-p port][-c nclients][-t ntransactions][-s scaling_factor][-n][-C][-v][-S][-N][-l][-U login][-P password][-d][dbname]\n");
 	fprintf(stderr, "(initialize mode): pgbench -i [-h hostname][-p port][-s scaling_factor][-U login][-P password][-d][dbname]\n");
@@ -118,7 +122,7 @@ getrand(int min, int max)
 
 /* set up a connection to the backend */
 static PGconn *
-doConnect()
+doConnect(void)
 {
 	PGconn	   *con;
 	PGresult   *res;
@@ -261,14 +265,14 @@ doOne(CState * state, int n, int debug, int ttype)
 				 */
 				if (use_log)
 				{
-					long long	diff;
+					double		diff;
 					struct timeval now;
 
-					gettimeofday(&now, 0);
-					diff = (now.tv_sec - st->txn_begin.tv_sec) * 1000000 +
-						(now.tv_usec - st->txn_begin.tv_usec);
+					gettimeofday(&now, NULL);
+					diff = (int) (now.tv_sec - st->txn_begin.tv_sec) * 1000000.0 +
+						(int) (now.tv_usec - st->txn_begin.tv_usec);
 
-					fprintf(LOGFILE, "%d %d %lld\n", st->id, st->cnt, diff);
+					fprintf(LOGFILE, "%d %d %.0f\n", st->id, st->cnt, diff);
 				}
 
 				res = PQgetResult(st->con);
@@ -324,7 +328,7 @@ doOne(CState * state, int n, int debug, int ttype)
 			st->tid = getrand(1, ntellers * tps);
 			st->delta = getrand(1, 1000);
 			if (use_log)
-				gettimeofday(&(st->txn_begin), 0);
+				gettimeofday(&(st->txn_begin), NULL);
 			break;
 		case 1:
 			snprintf(sql, 256, "update accounts set abalance = abalance + %d where aid = %d\n", st->delta, st->aid);
@@ -482,13 +486,19 @@ init(void)
 	PGresult   *res;
 	static char *DDLs[] = {
 		"drop table branches",
-		"create table branches(bid int, primary key(bid),bbalance int,filler char(88))",
+		"create table branches(bid int not null,bbalance int,filler char(88))",
 		"drop table tellers",
-		"create table tellers(tid int, primary key(tid),bid int,tbalance int,filler char(84))",
+		"create table tellers(tid int not null,bid int,tbalance int,filler char(84))",
 		"drop table accounts",
-		"create table accounts(aid int,primary key(aid),bid int,abalance int,filler char(84))",
+		"create table accounts(aid int not null,bid int,abalance int,filler char(84))",
 		"drop table history",
 	"create table history(tid int,bid int,aid int,delta int,mtime timestamp,filler char(22))"};
+	static char *DDLAFTERs[] = {
+		"alter table branches add primary key (bid)",
+		"alter table tellers add primary key (tid)",
+	"alter table accounts add primary key (aid)"};
+
+
 	char		sql[256];
 
 	int			i;
@@ -546,7 +556,7 @@ init(void)
 		fprintf(stderr, "%s", PQerrorMessage(con));
 		exit(1);
 	}
-
+	PQclear(res);
 
 	/*
 	 * occupy accounts table with some data
@@ -608,6 +618,17 @@ init(void)
 #endif   /* NOT_USED */
 		}
 	}
+	fprintf(stderr, "set primary key...\n");
+	for (i = 0; i < (sizeof(DDLAFTERs) / sizeof(char *)); i++)
+	{
+		res = PQexec(con, DDLAFTERs[i]);
+		if (strncmp(DDLs[i], "drop", 4) && PQresultStatus(res) != PGRES_COMMAND_OK)
+		{
+			fprintf(stderr, "%s", PQerrorMessage(con));
+			exit(1);
+		}
+		PQclear(res);
+	}
 
 	/* vacuum */
 	fprintf(stderr, "vacuum...");
@@ -688,7 +709,7 @@ main(int argc, char **argv)
 	int			nsocks;			/* return from select(2) */
 	int			maxsock;		/* max socket number to be waited */
 
-#ifndef __CYGWIN__
+#if !(defined(__CYGWIN__) || defined(__MINGW32__))
 	struct rlimit rlim;
 #endif
 
@@ -738,7 +759,7 @@ main(int argc, char **argv)
 					fprintf(stderr, "invalid number of clients: %d\n", nclients);
 					exit(1);
 				}
-#ifndef __CYGWIN__
+#if !(defined(__CYGWIN__) || defined(__MINGW32__))
 #ifdef RLIMIT_NOFILE			/* most platform uses RLIMIT_NOFILE */
 				if (getrlimit(RLIMIT_NOFILE, &rlim) == -1)
 				{
@@ -755,7 +776,7 @@ main(int argc, char **argv)
 					fprintf(stderr, "Use limit/ulimt to increase the limit before using pgbench.\n");
 					exit(1);
 				}
-#endif   /* #ifndef __CYGWIN__ */
+#endif   /* #if !(defined(__CYGWIN__) || defined(__MINGW32__)) */
 				break;
 			case 'C':
 				is_connect = 1;
@@ -917,11 +938,11 @@ main(int argc, char **argv)
 	PQfinish(con);
 
 	/* set random seed */
-	gettimeofday(&tv1, 0);
-	srand((uint) tv1.tv_usec);
+	gettimeofday(&tv1, NULL);
+	srand((unsigned int) tv1.tv_usec);
 
 	/* get start up time */
-	gettimeofday(&tv1, 0);
+	gettimeofday(&tv1, NULL);
 
 	if (is_connect == 0)
 	{
@@ -935,7 +956,7 @@ main(int argc, char **argv)
 	}
 
 	/* time after connections set up */
-	gettimeofday(&tv2, 0);
+	gettimeofday(&tv2, NULL);
 
 	/* send start up queries in async manner */
 	for (i = 0; i < nclients; i++)
@@ -952,7 +973,7 @@ main(int argc, char **argv)
 		{						/* all done ? */
 			disconnect_all(state);
 			/* get end time */
-			gettimeofday(&tv3, 0);
+			gettimeofday(&tv3, NULL);
 			printResults(ttype, state, &tv1, &tv2, &tv3);
 			if (LOGFILE)
 				fclose(LOGFILE);
