@@ -3,12 +3,12 @@
  * fmgr.c
  *	  The Postgres function manager.
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/utils/fmgr/fmgr.c,v 1.63 2002/10/04 17:19:55 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/fmgr/fmgr.c,v 1.88 2004/12/31 22:01:31 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -20,6 +20,7 @@
 #include "catalog/pg_proc.h"
 #include "executor/functions.h"
 #include "miscadmin.h"
+#include "parser/parse_expr.h"
 #include "utils/builtins.h"
 #include "utils/fmgrtab.h"
 #include "utils/lsyscache.h"
@@ -33,17 +34,16 @@
  * %d0. So if a function pointer is declared to return a pointer, the
  * compiler may look only into %a0, but if the called function was declared
  * to return an integer type, it puts its value only into %d0. So the
- * caller doesn't pink up the correct return value. The solution is to
+ * caller doesn't pick up the correct return value. The solution is to
  * declare the function pointer to return int, so the compiler picks up the
  * return value from %d0. (Functions returning pointers put their value
  * *additionally* into %d0 for compatibility.) The price is that there are
  * some warnings about int->pointer conversions...
  */
-#if defined(__mc68000__) && defined(__ELF__)
-typedef int32 ((*func_ptr) ());
-
+#if (defined(__mc68000__) || (defined(__m68k__))) && defined(__ELF__)
+typedef int32  (*func_ptr) ();
 #else
-typedef char *((*func_ptr) ());
+typedef char * (*func_ptr) ();
 #endif
 
 /*
@@ -56,11 +56,29 @@ typedef struct
 												 * toastable datatype? */
 } Oldstyle_fnextra;
 
+/*
+ * Hashtable for fast lookup of external C functions
+ */
+typedef struct
+{
+	/* fn_oid is the hash key and so must be first! */
+	Oid			fn_oid;			/* OID of an external C function */
+	TransactionId fn_xmin;		/* for checking up-to-dateness */
+	CommandId	fn_cmin;
+	PGFunction	user_fn;		/* the function's address */
+	Pg_finfo_record *inforec;	/* address of its info record */
+} CFuncHashTabEntry;
+
+static HTAB *CFuncHash = NULL;
+
 
 static void fmgr_info_cxt_security(Oid functionId, FmgrInfo *finfo, MemoryContext mcxt,
 					   bool ignore_security);
 static void fmgr_info_C_lang(Oid functionId, FmgrInfo *finfo, HeapTuple procedureTuple);
 static void fmgr_info_other_lang(Oid functionId, FmgrInfo *finfo, HeapTuple procedureTuple);
+static CFuncHashTabEntry *lookup_C_func(HeapTuple procedureTuple);
+static void record_C_func(HeapTuple procedureTuple,
+			  PGFunction user_fn, Pg_finfo_record *inforec);
 static Datum fmgr_oldstyle(PG_FUNCTION_ARGS);
 static Datum fmgr_security_definer(PG_FUNCTION_ARGS);
 
@@ -92,7 +110,7 @@ fmgr_isbuiltin(Oid id)
 		else
 			high = i - 1;
 	}
-	return (const FmgrBuiltin *) NULL;
+	return NULL;
 }
 
 /*
@@ -110,7 +128,7 @@ fmgr_lookupByName(const char *name)
 		if (strcmp(name, fmgr_builtins[i].funcName) == 0)
 			return fmgr_builtins + i;
 	}
-	return (const FmgrBuiltin *) NULL;
+	return NULL;
 }
 
 /*
@@ -154,6 +172,8 @@ fmgr_info_cxt_security(Oid functionId, FmgrInfo *finfo, MemoryContext mcxt,
 	const FmgrBuiltin *fbp;
 	HeapTuple	procedureTuple;
 	Form_pg_proc procedureStruct;
+	Datum		prosrcdatum;
+	bool		isnull;
 	char	   *prosrc;
 
 	/*
@@ -164,6 +184,7 @@ fmgr_info_cxt_security(Oid functionId, FmgrInfo *finfo, MemoryContext mcxt,
 	finfo->fn_oid = InvalidOid;
 	finfo->fn_extra = NULL;
 	finfo->fn_mcxt = mcxt;
+	finfo->fn_expr = NULL;		/* caller may set this later */
 
 	if ((fbp = fmgr_isbuiltin(functionId)) != NULL)
 	{
@@ -184,8 +205,7 @@ fmgr_info_cxt_security(Oid functionId, FmgrInfo *finfo, MemoryContext mcxt,
 									ObjectIdGetDatum(functionId),
 									0, 0, 0);
 	if (!HeapTupleIsValid(procedureTuple))
-		elog(ERROR, "fmgr_info: function %u: cache lookup failed",
-			 functionId);
+		elog(ERROR, "cache lookup failed for function %u", functionId);
 	procedureStruct = (Form_pg_proc) GETSTRUCT(procedureTuple);
 
 	finfo->fn_nargs = procedureStruct->pronargs;
@@ -213,12 +233,18 @@ fmgr_info_cxt_security(Oid functionId, FmgrInfo *finfo, MemoryContext mcxt,
 			 * name of the internal function is stored in prosrc (it
 			 * doesn't have to be the same as the name of the alias!)
 			 */
+			prosrcdatum = SysCacheGetAttr(PROCOID, procedureTuple,
+										  Anum_pg_proc_prosrc, &isnull);
+			if (isnull)
+				elog(ERROR, "null prosrc");
 			prosrc = DatumGetCString(DirectFunctionCall1(textout,
-							 PointerGetDatum(&procedureStruct->prosrc)));
+														 prosrcdatum));
 			fbp = fmgr_lookupByName(prosrc);
 			if (fbp == NULL)
-				elog(ERROR, "fmgr_info: function %s not in internal table",
-					 prosrc);
+				ereport(ERROR,
+						(errcode(ERRCODE_UNDEFINED_FUNCTION),
+						 errmsg("internal function \"%s\" is not in internal lookup table",
+								prosrc)));
 			pfree(prosrc);
 			/* Should we check that nargs, strict, retset match the table? */
 			finfo->fn_addr = fbp->func;
@@ -249,38 +275,61 @@ static void
 fmgr_info_C_lang(Oid functionId, FmgrInfo *finfo, HeapTuple procedureTuple)
 {
 	Form_pg_proc procedureStruct = (Form_pg_proc) GETSTRUCT(procedureTuple);
-	Datum		prosrcattr,
-				probinattr;
-	char	   *prosrcstring,
-			   *probinstring;
-	void	   *libraryhandle;
+	CFuncHashTabEntry *hashentry;
 	PGFunction	user_fn;
 	Pg_finfo_record *inforec;
 	Oldstyle_fnextra *fnextra;
 	bool		isnull;
 	int			i;
 
-	/* Get prosrc and probin strings (link symbol and library filename) */
-	prosrcattr = SysCacheGetAttr(PROCOID, procedureTuple,
-								 Anum_pg_proc_prosrc, &isnull);
-	if (isnull)
-		elog(ERROR, "fmgr: Could not extract prosrc for %u from pg_proc",
-			 functionId);
-	prosrcstring = DatumGetCString(DirectFunctionCall1(textout, prosrcattr));
+	/*
+	 * See if we have the function address cached already
+	 */
+	hashentry = lookup_C_func(procedureTuple);
+	if (hashentry)
+	{
+		user_fn = hashentry->user_fn;
+		inforec = hashentry->inforec;
+	}
+	else
+	{
+		Datum		prosrcattr,
+					probinattr;
+		char	   *prosrcstring,
+				   *probinstring;
+		void	   *libraryhandle;
 
-	probinattr = SysCacheGetAttr(PROCOID, procedureTuple,
-								 Anum_pg_proc_probin, &isnull);
-	if (isnull)
-		elog(ERROR, "fmgr: Could not extract probin for %u from pg_proc",
-			 functionId);
-	probinstring = DatumGetCString(DirectFunctionCall1(textout, probinattr));
+		/*
+		 * Get prosrc and probin strings (link symbol and library
+		 * filename)
+		 */
+		prosrcattr = SysCacheGetAttr(PROCOID, procedureTuple,
+									 Anum_pg_proc_prosrc, &isnull);
+		if (isnull)
+			elog(ERROR, "null prosrc for function %u", functionId);
+		prosrcstring = DatumGetCString(DirectFunctionCall1(textout,
+														   prosrcattr));
 
-	/* Look up the function itself */
-	user_fn = load_external_function(probinstring, prosrcstring, true,
-									 &libraryhandle);
+		probinattr = SysCacheGetAttr(PROCOID, procedureTuple,
+									 Anum_pg_proc_probin, &isnull);
+		if (isnull)
+			elog(ERROR, "null probin for function %u", functionId);
+		probinstring = DatumGetCString(DirectFunctionCall1(textout,
+														   probinattr));
 
-	/* Get the function information record (real or default) */
-	inforec = fetch_finfo_record(libraryhandle, prosrcstring);
+		/* Look up the function itself */
+		user_fn = load_external_function(probinstring, prosrcstring, true,
+										 &libraryhandle);
+
+		/* Get the function information record (real or default) */
+		inforec = fetch_finfo_record(libraryhandle, prosrcstring);
+
+		/* Cache the addresses for later calls */
+		record_C_func(procedureTuple, user_fn, inforec);
+
+		pfree(prosrcstring);
+		pfree(probinstring);
+	}
 
 	switch (inforec->api_version)
 	{
@@ -304,13 +353,10 @@ fmgr_info_C_lang(Oid functionId, FmgrInfo *finfo, HeapTuple procedureTuple)
 			break;
 		default:
 			/* Shouldn't get here if fetch_finfo_record did its job */
-			elog(ERROR, "Unknown function API version %d",
+			elog(ERROR, "unrecognized function API version: %d",
 				 inforec->api_version);
 			break;
 	}
-
-	pfree(prosrcstring);
-	pfree(probinstring);
 }
 
 /*
@@ -330,8 +376,7 @@ fmgr_info_other_lang(Oid functionId, FmgrInfo *finfo, HeapTuple procedureTuple)
 								   ObjectIdGetDatum(language),
 								   0, 0, 0);
 	if (!HeapTupleIsValid(languageTuple))
-		elog(ERROR, "fmgr_info: cache lookup for language %u failed",
-			 language);
+		elog(ERROR, "cache lookup failed for language %u", language);
 	languageStruct = (Form_pg_language) GETSTRUCT(languageTuple);
 
 	fmgr_info(languageStruct->lanplcallfoid, &plfinfo);
@@ -343,8 +388,7 @@ fmgr_info_other_lang(Oid functionId, FmgrInfo *finfo, HeapTuple procedureTuple)
 	 * oldstyle PL handlers.
 	 */
 	if (plfinfo.fn_extra != NULL)
-		elog(ERROR, "fmgr_info: language %u has old-style handler",
-			 language);
+		elog(ERROR, "language %u has old-style handler", language);
 
 	ReleaseSysCache(languageTuple);
 }
@@ -379,7 +423,7 @@ fetch_finfo_record(void *filehandle, char *funcname)
 	/* Try to look up the info function */
 	infofunc = (PGFInfoFunction) lookup_external_function(filehandle,
 														  infofuncname);
-	if (infofunc == (PGFInfoFunction) NULL)
+	if (infofunc == NULL)
 	{
 		/* Not found --- assume version 0 */
 		pfree(infofuncname);
@@ -391,7 +435,7 @@ fetch_finfo_record(void *filehandle, char *funcname)
 
 	/* Validate result as best we can */
 	if (inforec == NULL)
-		elog(ERROR, "Null result from %s", infofuncname);
+		elog(ERROR, "null result from info function \"%s\"", infofuncname);
 	switch (inforec->api_version)
 	{
 		case 0:
@@ -399,8 +443,10 @@ fetch_finfo_record(void *filehandle, char *funcname)
 			/* OK, no additional fields to validate */
 			break;
 		default:
-			elog(ERROR, "Unknown version %d reported by %s",
-				 inforec->api_version, infofuncname);
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("unrecognized API version %d reported by info function \"%s\"",
+							inforec->api_version, infofuncname)));
 			break;
 	}
 
@@ -409,6 +455,98 @@ fetch_finfo_record(void *filehandle, char *funcname)
 }
 
 
+/*-------------------------------------------------------------------------
+ *		Routines for caching lookup information for external C functions.
+ *
+ * The routines in dfmgr.c are relatively slow, so we try to avoid running
+ * them more than once per external function per session.  We use a hash table
+ * with the function OID as the lookup key.
+ *-------------------------------------------------------------------------
+ */
+
+/*
+ * lookup_C_func: try to find a C function in the hash table
+ *
+ * If an entry exists and is up to date, return it; else return NULL
+ */
+static CFuncHashTabEntry *
+lookup_C_func(HeapTuple procedureTuple)
+{
+	Oid			fn_oid = HeapTupleGetOid(procedureTuple);
+	CFuncHashTabEntry *entry;
+
+	if (CFuncHash == NULL)
+		return NULL;			/* no table yet */
+	entry = (CFuncHashTabEntry *)
+		hash_search(CFuncHash,
+					&fn_oid,
+					HASH_FIND,
+					NULL);
+	if (entry == NULL)
+		return NULL;			/* no such entry */
+	if (entry->fn_xmin == HeapTupleHeaderGetXmin(procedureTuple->t_data) &&
+		entry->fn_cmin == HeapTupleHeaderGetCmin(procedureTuple->t_data))
+		return entry;			/* OK */
+	return NULL;				/* entry is out of date */
+}
+
+/*
+ * record_C_func: enter (or update) info about a C function in the hash table
+ */
+static void
+record_C_func(HeapTuple procedureTuple,
+			  PGFunction user_fn, Pg_finfo_record *inforec)
+{
+	Oid			fn_oid = HeapTupleGetOid(procedureTuple);
+	CFuncHashTabEntry *entry;
+	bool		found;
+
+	/* Create the hash table if it doesn't exist yet */
+	if (CFuncHash == NULL)
+	{
+		HASHCTL		hash_ctl;
+
+		MemSet(&hash_ctl, 0, sizeof(hash_ctl));
+		hash_ctl.keysize = sizeof(Oid);
+		hash_ctl.entrysize = sizeof(CFuncHashTabEntry);
+		hash_ctl.hash = tag_hash;
+		CFuncHash = hash_create("CFuncHash",
+								100,
+								&hash_ctl,
+								HASH_ELEM | HASH_FUNCTION);
+	}
+
+	entry = (CFuncHashTabEntry *)
+		hash_search(CFuncHash,
+					&fn_oid,
+					HASH_ENTER,
+					&found);
+	if (entry == NULL)
+		ereport(ERROR,
+				(errcode(ERRCODE_OUT_OF_MEMORY),
+				 errmsg("out of memory")));
+	/* OID is already filled in */
+	entry->fn_xmin = HeapTupleHeaderGetXmin(procedureTuple->t_data);
+	entry->fn_cmin = HeapTupleHeaderGetCmin(procedureTuple->t_data);
+	entry->user_fn = user_fn;
+	entry->inforec = inforec;
+}
+
+/*
+ * clear_external_function_hash: remove entries for a library being closed
+ *
+ * Presently we just zap the entire hash table, but later it might be worth
+ * the effort to remove only the entries associated with the given handle.
+ */
+void
+clear_external_function_hash(void *filehandle)
+{
+	if (CFuncHash)
+		hash_destroy(CFuncHash);
+	CFuncHash = NULL;
+}
+
+
 /*
  * Copy an FmgrInfo struct
  *
@@ -467,7 +605,7 @@ fmgr_oldstyle(PG_FUNCTION_ARGS)
 	char	   *returnValue;
 
 	if (fcinfo->flinfo == NULL || fcinfo->flinfo->fn_extra == NULL)
-		elog(ERROR, "Internal error: fmgr_oldstyle received NULL pointer");
+		elog(ERROR, "fmgr_oldstyle received NULL pointer");
 	fnextra = (Oldstyle_fnextra *) fcinfo->flinfo->fn_extra;
 
 	/*
@@ -616,8 +754,10 @@ fmgr_oldstyle(PG_FUNCTION_ARGS)
 			 * needed to support old-style functions with many arguments,
 			 * but making 'em be new-style is probably a better idea.
 			 */
-			elog(ERROR, "fmgr_oldstyle: function %u: too many arguments (%d > %d)",
-				 fcinfo->flinfo->fn_oid, n_arguments, 16);
+			ereport(ERROR,
+					(errcode(ERRCODE_TOO_MANY_ARGUMENTS),
+					 errmsg("function %u has too many arguments (%d, maximum is %d)",
+							fcinfo->flinfo->fn_oid, n_arguments, 16)));
 			returnValue = NULL; /* keep compiler quiet */
 			break;
 	}
@@ -633,7 +773,7 @@ fmgr_oldstyle(PG_FUNCTION_ARGS)
 struct fmgr_security_definer_cache
 {
 	FmgrInfo	flinfo;
-	Oid			userid;
+	AclId		userid;
 };
 
 /*
@@ -649,8 +789,8 @@ fmgr_security_definer(PG_FUNCTION_ARGS)
 {
 	Datum		result;
 	FmgrInfo   *save_flinfo;
-	struct fmgr_security_definer_cache *fcache;
-	Oid			save_userid;
+	struct fmgr_security_definer_cache * volatile fcache;
+	AclId		save_userid;
 	HeapTuple	tuple;
 
 	if (!fcinfo->flinfo->fn_extra)
@@ -661,9 +801,11 @@ fmgr_security_definer(PG_FUNCTION_ARGS)
 		fmgr_info_cxt_security(fcinfo->flinfo->fn_oid, &fcache->flinfo,
 							   fcinfo->flinfo->fn_mcxt, true);
 
-		tuple = SearchSysCache(PROCOID, ObjectIdGetDatum(fcinfo->flinfo->fn_oid), 0, 0, 0);
+		tuple = SearchSysCache(PROCOID,
+							   ObjectIdGetDatum(fcinfo->flinfo->fn_oid),
+							   0, 0, 0);
 		if (!HeapTupleIsValid(tuple))
-			elog(ERROR, "fmgr_security_definer: function %u: cache lookup failed",
+			elog(ERROR, "cache lookup failed for function %u",
 				 fcinfo->flinfo->fn_oid);
 		fcache->userid = ((Form_pg_proc) GETSTRUCT(tuple))->proowner;
 		ReleaseSysCache(tuple);
@@ -674,14 +816,25 @@ fmgr_security_definer(PG_FUNCTION_ARGS)
 		fcache = fcinfo->flinfo->fn_extra;
 
 	save_flinfo = fcinfo->flinfo;
-	fcinfo->flinfo = &fcache->flinfo;
-
 	save_userid = GetUserId();
-	SetUserId(fcache->userid);
-	result = FunctionCallInvoke(fcinfo);
-	SetUserId(save_userid);
+
+	PG_TRY();
+	{
+		fcinfo->flinfo = &fcache->flinfo;
+		SetUserId(fcache->userid);
+
+		result = FunctionCallInvoke(fcinfo);
+	}
+	PG_CATCH();
+	{
+		fcinfo->flinfo = save_flinfo;
+		SetUserId(save_userid);
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
 
 	fcinfo->flinfo = save_flinfo;
+	SetUserId(save_userid);
 
 	return result;
 }
@@ -693,7 +846,7 @@ fmgr_security_definer(PG_FUNCTION_ARGS)
  * NOTE: the simplest way to reliably initialize a FunctionCallInfoData
  * is to MemSet it to zeroes and then fill in the fields that should be
  * nonzero.  However, in a few of the most heavily used paths, we instead
- * just zero the fields that must be zero.  This saves a fair number of
+ * just zero the fields that must be zero.	This saves a fair number of
  * cycles so it's worth the extra maintenance effort.  Also see inlined
  * version of FunctionCall2 in utils/sort/tuplesort.c if you need to change
  * these routines!
@@ -725,8 +878,7 @@ DirectFunctionCall1(PGFunction func, Datum arg1)
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall1: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -753,8 +905,7 @@ DirectFunctionCall2(PGFunction func, Datum arg1, Datum arg2)
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall2: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -776,8 +927,7 @@ DirectFunctionCall3(PGFunction func, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall3: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -800,8 +950,7 @@ DirectFunctionCall4(PGFunction func, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall4: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -825,8 +974,7 @@ DirectFunctionCall5(PGFunction func, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall5: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -852,8 +1000,7 @@ DirectFunctionCall6(PGFunction func, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall6: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -880,8 +1027,7 @@ DirectFunctionCall7(PGFunction func, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall7: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -909,8 +1055,7 @@ DirectFunctionCall8(PGFunction func, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall8: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -940,8 +1085,7 @@ DirectFunctionCall9(PGFunction func, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "DirectFunctionCall9: function %p returned NULL",
-			 (void *) func);
+		elog(ERROR, "function %p returned NULL", (void *) func);
 
 	return result;
 }
@@ -971,8 +1115,7 @@ FunctionCall1(FmgrInfo *flinfo, Datum arg1)
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall1: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -999,8 +1142,7 @@ FunctionCall2(FmgrInfo *flinfo, Datum arg1, Datum arg2)
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall2: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -1023,8 +1165,7 @@ FunctionCall3(FmgrInfo *flinfo, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall3: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -1048,8 +1189,7 @@ FunctionCall4(FmgrInfo *flinfo, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall4: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -1074,8 +1214,7 @@ FunctionCall5(FmgrInfo *flinfo, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall5: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -1102,8 +1241,7 @@ FunctionCall6(FmgrInfo *flinfo, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall6: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -1131,8 +1269,7 @@ FunctionCall7(FmgrInfo *flinfo, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall7: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -1161,8 +1298,7 @@ FunctionCall8(FmgrInfo *flinfo, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall8: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -1193,8 +1329,7 @@ FunctionCall9(FmgrInfo *flinfo, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "FunctionCall9: function %u returned NULL",
-			 fcinfo.flinfo->fn_oid);
+		elog(ERROR, "function %u returned NULL", fcinfo.flinfo->fn_oid);
 
 	return result;
 }
@@ -1224,8 +1359,7 @@ OidFunctionCall1(Oid functionId, Datum arg1)
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall1: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1249,8 +1383,7 @@ OidFunctionCall2(Oid functionId, Datum arg1, Datum arg2)
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall2: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1276,8 +1409,7 @@ OidFunctionCall3(Oid functionId, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall3: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1304,8 +1436,7 @@ OidFunctionCall4(Oid functionId, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall4: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1333,8 +1464,7 @@ OidFunctionCall5(Oid functionId, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall5: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1364,8 +1494,7 @@ OidFunctionCall6(Oid functionId, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall6: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1396,8 +1525,7 @@ OidFunctionCall7(Oid functionId, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall7: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1429,8 +1557,7 @@ OidFunctionCall8(Oid functionId, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall8: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1464,8 +1591,7 @@ OidFunctionCall9(Oid functionId, Datum arg1, Datum arg2,
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "OidFunctionCall9: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return result;
 }
@@ -1503,8 +1629,10 @@ fmgr(Oid procedureId,...)
 		int			i;
 
 		if (n_arguments > FUNC_MAX_ARGS)
-			elog(ERROR, "fmgr: function %u: too many arguments (%d > %d)",
-				 flinfo.fn_oid, n_arguments, FUNC_MAX_ARGS);
+			ereport(ERROR,
+					(errcode(ERRCODE_TOO_MANY_ARGUMENTS),
+					 errmsg("function %u has too many arguments (%d, maximum is %d)",
+							flinfo.fn_oid, n_arguments, FUNC_MAX_ARGS)));
 		va_start(pvar, procedureId);
 		for (i = 0; i < n_arguments; i++)
 			fcinfo.arg[i] = (Datum) va_arg(pvar, char *);
@@ -1515,8 +1643,7 @@ fmgr(Oid procedureId,...)
 
 	/* Check for null result, since caller is clearly not expecting one */
 	if (fcinfo.isnull)
-		elog(ERROR, "fmgr: function %u returned NULL",
-			 flinfo.fn_oid);
+		elog(ERROR, "function %u returned NULL", flinfo.fn_oid);
 
 	return (char *) result;
 }
@@ -1550,9 +1677,8 @@ Int64GetDatum(int64 X)
 	 * with zeroes in the unused bits.	This is needed so that, for
 	 * example, hash join of int8 will behave properly.
 	 */
-	int64	   *retval = (int64 *) palloc(Max(sizeof(int64), 8));
+	int64	   *retval = (int64 *) palloc0(Max(sizeof(int64), 8));
 
-	MemSet(retval, 0, Max(sizeof(int64), 8));
 	*retval = X;
 	return PointerGetDatum(retval);
 #endif   /* INT64_IS_BUSTED */
@@ -1612,3 +1738,83 @@ pg_detoast_datum_slice(struct varlena * datum, int32 first, int32 count)
 	/* Only get the specified portion from the toast rel */
 	return (struct varlena *) heap_tuple_untoast_attr_slice((varattrib *) datum, first, count);
 }
+
+/*-------------------------------------------------------------------------
+ *		Support routines for extracting info from fn_expr parse tree
+ *
+ * These are needed by polymorphic functions, which accept multiple possible
+ * input types and need help from the parser to know what they've got.
+ *-------------------------------------------------------------------------
+ */
+
+/*
+ * Get the actual type OID of the function return type
+ *
+ * Returns InvalidOid if information is not available
+ */
+Oid
+get_fn_expr_rettype(FmgrInfo *flinfo)
+{
+	Node	   *expr;
+
+	/*
+	 * can't return anything useful if we have no FmgrInfo or if its
+	 * fn_expr node has not been initialized
+	 */
+	if (!flinfo || !flinfo->fn_expr)
+		return InvalidOid;
+
+	expr = flinfo->fn_expr;
+
+	return exprType(expr);
+}
+
+/*
+ * Get the actual type OID of a specific function argument (counting from 0)
+ *
+ * Returns InvalidOid if information is not available
+ */
+Oid
+get_fn_expr_argtype(FmgrInfo *flinfo, int argnum)
+{
+	Node	   *expr;
+	List	   *args;
+	Oid			argtype;
+
+	/*
+	 * can't return anything useful if we have no FmgrInfo or if its
+	 * fn_expr node has not been initialized
+	 */
+	if (!flinfo || !flinfo->fn_expr)
+		return InvalidOid;
+
+	expr = flinfo->fn_expr;
+
+	if (IsA(expr, FuncExpr))
+		args = ((FuncExpr *) expr)->args;
+	else if (IsA(expr, OpExpr))
+		args = ((OpExpr *) expr)->args;
+	else if (IsA(expr, DistinctExpr))
+		args = ((DistinctExpr *) expr)->args;
+	else if (IsA(expr, ScalarArrayOpExpr))
+		args = ((ScalarArrayOpExpr *) expr)->args;
+	else if (IsA(expr, NullIfExpr))
+		args = ((NullIfExpr *) expr)->args;
+	else
+		return InvalidOid;
+
+	if (argnum < 0 || argnum >= list_length(args))
+		return InvalidOid;
+
+	argtype = exprType((Node *) list_nth(args, argnum));
+
+	/*
+	 * special hack for ScalarArrayOpExpr: what the underlying function
+	 * will actually get passed is the element type of the array.
+	 */
+	if (IsA(expr, ScalarArrayOpExpr) &&
+		argnum == 1)
+		argtype = get_element_type(argtype);
+
+	return argtype;
+}
